@@ -28,11 +28,13 @@ import asyncio
 import struct
 import math
 import copy
+from typing import Optional
 
-from autobahn.util import public, _LazyHexFormatter
-from autobahn.wamp.exception import ProtocolError, SerializationError, TransportLost
-from autobahn.asyncio.util import peer2str, get_serializers
 import txaio
+from autobahn.util import public, _LazyHexFormatter, hltype
+from autobahn.wamp.exception import ProtocolError, SerializationError, TransportLost
+from autobahn.wamp.types import TransportDetails
+from autobahn.asyncio.util import get_serializers, create_transport_details, transport_channel_id
 
 __all__ = (
     'WampRawSocketServerProtocol',
@@ -41,8 +43,6 @@ __all__ = (
     'WampRawSocketClientFactory'
 )
 
-txaio.use_asyncio()
-
 FRAME_TYPE_DATA = 0
 FRAME_TYPE_PING = 1
 FRAME_TYPE_PONG = 2
@@ -57,12 +57,30 @@ class PrefixProtocol(asyncio.Protocol):
     max_length = 16 * 1024 * 1024
     max_length_send = max_length
     log = txaio.make_logger()  # @UndefinedVariable
+    peer: Optional[str] = None
+    is_server: Optional[bool] = None
+
+    @property
+    def transport_details(self) -> Optional[TransportDetails]:
+        """
+        Implements :func:`autobahn.wamp.interfaces.ITransport.transport_details`
+        """
+        return self._transport_details
 
     def connection_made(self, transport):
-        self.transport = transport
-        peer = transport.get_extra_info('peername')
-        self.peer = peer2str(peer)
+        # asyncio networking framework entry point, called by asyncio
+        # when the connection is established (either a client or a server)
         self.log.debug('RawSocker Asyncio: Connection made with peer {peer}', peer=self.peer)
+
+        self.transport = transport
+
+        # determine preliminary transport details (what is known at this point)
+        self._transport_details = create_transport_details(self.transport, self.is_server)
+        self._transport_details.channel_framing = TransportDetails.CHANNEL_FRAMING_RAWSOCKET
+
+        # backward compatibility
+        self.peer = self._transport_details.peer
+
         self._buffer = b''
         self._header = None
         self._wait_closed = txaio.create_future()
@@ -172,7 +190,6 @@ class RawSocketProtocol(PrefixProtocol):
         buf = bytearray(self._buffer[:4])
         if buf[0] != MAGIC_BYTE:
             raise HandshakeError('Invalid magic byte in handshake')
-            return
         ser = buf[1] & 0x0F
         lexp = buf[1] >> 4
         self.max_length_send = 2**(lexp + 9)
@@ -221,6 +238,8 @@ class HandshakeError(Exception):
 
 class RawSocketClientProtocol(RawSocketProtocol):
 
+    is_server = False
+
     def check_serializer(self, ser_id):
         return True
 
@@ -248,6 +267,8 @@ class RawSocketClientProtocol(RawSocketProtocol):
 
 class RawSocketServerProtocol(RawSocketProtocol):
 
+    is_server = True
+
     def supports_serializer(self, ser_id):
         raise NotImplementedError()
 
@@ -267,7 +288,17 @@ class WampRawSocketMixinGeneral(object):
 
     def _on_handshake_complete(self):
         self.log.debug("WampRawSocketProtocol: Handshake complete")
+        # RawSocket connection established. Now let the user WAMP session factory
+        # create a new WAMP session and fire off session open callback.
         try:
+            if self._transport_details.is_secure:
+                # now that the TLS opening handshake is complete, the actual TLS channel ID
+                # will be available. make sure to set it!
+                channel_id = {
+                    'tls-unique': transport_channel_id(self.transport, self._transport_details.is_server, 'tls-unique'),
+                }
+                self._transport_details.channel_id = channel_id
+
             self._session = self.factory._factory()
             self._session.onOpen(self)
         except Exception as e:
@@ -297,7 +328,7 @@ class WampRawSocketMixinGeneral(object):
         Implements :func:`autobahn.wamp.interfaces.ITransport.send`
         """
         if self.isOpen():
-            self.log.debug("WampRawSocketProtocol: TX WAMP message: {msg}", msg=msg)
+            self.log.debug('{func}: TX WAMP message: {msg}', func=hltype(self.send), msg=msg)
             try:
                 payload, _ = self._serializer.serialize(msg)
             except Exception as e:
@@ -382,13 +413,6 @@ class WampRawSocketServerProtocol(WampRawSocketMixinGeneral, WampRawSocketMixinA
             self.abort()
             return False
 
-    def get_channel_id(self, channel_id_type='tls-unique'):
-        """
-        Implements :func:`autobahn.wamp.interfaces.ITransport.get_channel_id`
-        """
-        return None
-        # return transport_channel_id(self.transport, is_server=True, channel_id_type=channel_id_type)
-
 
 @public
 class WampRawSocketClientProtocol(WampRawSocketMixinGeneral, WampRawSocketMixinAsyncio, RawSocketClientProtocol):
@@ -406,13 +430,6 @@ class WampRawSocketClientProtocol(WampRawSocketMixinGeneral, WampRawSocketMixinA
             self._serializer = copy.copy(self.factory._serializer)
         return self._serializer.RAWSOCKET_SERIALIZER_ID
 
-    def get_channel_id(self, channel_id_type='tls-unique'):
-        """
-        Implements :func:`autobahn.wamp.interfaces.ITransport.get_channel_id`
-        """
-        return None
-        # return transport_channel_id(self.transport, is_server=False, channel_id_type=channel_id_type)
-
 
 class WampRawSocketFactory(object):
     """
@@ -455,7 +472,7 @@ class WampRawSocketServerFactory(WampRawSocketFactory):
         # when no serializers were requested specifically, then support
         # all that are available
         if serializers is None:
-            serializers = get_serializers()
+            serializers = [serializer() for serializer in get_serializers()]
 
             if not serializers:
                 raise Exception("could not import any WAMP serializers")
