@@ -50,6 +50,12 @@ try:
     from autobahn.xbr._abi import XBR_DEBUG_TOKEN_ADDR_SRC, XBR_DEBUG_NETWORK_ADDR_SRC, XBR_DEBUG_MARKET_ADDR_SRC, XBR_DEBUG_CATALOG_ADDR_SRC, XBR_DEBUG_CHANNEL_ADDR_SRC  # noqa
     from autobahn.xbr._interfaces import IMarketMaker, IProvider, IConsumer, ISeller, IBuyer, IDelegate  # noqa
     from autobahn.xbr._util import make_w3, pack_uint256, unpack_uint256  # noqa
+    from autobahn.xbr._eip712_certificate import EIP712Certificate  # noqa
+    from autobahn.xbr._eip712_certificate_chain import parse_certificate_chain  # noqa
+    from autobahn.xbr._eip712_authority_certificate import sign_eip712_authority_certificate, \
+        recover_eip712_authority_certificate, create_eip712_authority_certificate, EIP712AuthorityCertificate  # noqa
+    from autobahn.xbr._eip712_delegate_certificate import sign_eip712_delegate_certificate, \
+        recover_eip712_delegate_certificate, create_eip712_delegate_certificate, EIP712DelegateCertificate  # noqa
     from autobahn.xbr._eip712_member_register import sign_eip712_member_register, recover_eip712_member_register  # noqa
     from autobahn.xbr._eip712_member_login import sign_eip712_member_login, recover_eip712_member_login  # noqa
     from autobahn.xbr._eip712_market_create import sign_eip712_market_create, recover_eip712_market_create  # noqa
@@ -70,12 +76,12 @@ try:
     from autobahn.xbr._schema import FbsSchema, FbsObject, FbsType, FbsRPCCall, FbsEnum, FbsService, FbsEnumValue, \
     FbsAttribute, FbsField, FbsRepository  # noqa
     from autobahn.xbr._wallet import stretch_argon2_secret, expand_argon2_secret, pkm_from_argon2_secret  # noqa
+    from autobahn.xbr._frealm import FederatedRealm, Seeder  # noqa
+    from autobahn.xbr._secmod import EthereumKey, SecurityModuleMemory  # noqa
+    from autobahn.xbr._userkey import UserKey  # noqa
 
     HAS_XBR = True
 
-    from autobahn.xbr._frealm import FederatedRealm, Seeder  # noqa
-    from autobahn.xbr._secmod import EthereumKey  # noqa
-
     if not hasattr(abi, 'collapse_type'):
 
         def collapse_type(base, sub, arrlist):
@@ -327,6 +333,18 @@ try:
         'account_from_seedphrase',
         'ASCII_BOMB',
 
+        'EIP712Certificate',
+        'EIP712AuthorityCertificate',
+        'EIP712DelegateCertificate',
+        'parse_certificate_chain',
+
+        'create_eip712_authority_certificate',
+        'sign_eip712_authority_certificate',
+        'recover_eip712_authority_certificate',
+        'create_eip712_delegate_certificate',
+        'sign_eip712_delegate_certificate',
+        'recover_eip712_delegate_certificate',
+
         'sign_eip712_member_register',
         'recover_eip712_member_register',
         'sign_eip712_member_login',
@@ -359,6 +377,7 @@ try:
         'load_or_create_profile',
         'UserConfig',
         'Profile',
+        'UserKey',
 
         'MemberLevel',
         'ActorType',
@@ -377,6 +396,7 @@ try:
         'IBuyer',
         'IDelegate',
 
+        'FbsRepository',
         'FbsSchema',
         'FbsService',
         'FbsType',
@@ -393,6 +413,7 @@ try:
         'FederatedRealm',
         'Seeder',
         'EthereumKey',
+        'SecurityModuleMemory',
     )
 
 except (ImportError, FileNotFoundError) as e:
